/* GTK - The GIMP Toolkit
 * Copyright (C) 2002 Jeff Franks <jcf@tpg.com.au>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkdockbox.h"
#include <gtk/gtkenums.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkhbox.h>

/* Temporary define so that _(X) will compile. Will fix later. */

#define _(X) X

/* GtkDockBar */

enum {
  PROP_0,
  PROP_POSITION,
};

enum {
  ORIENTATION_CHANGED,
  POSITION_CHANGED,
  LAST_SIGNAL
};

#define DRAG_HANDLE_SIZE 10

static void       gtk_dock_bar_class_init               (GtkDockBarClass *klass);
static void       gtk_dock_bar_init                     (GtkDockBar      *box);

static void       gtk_dock_bar_set_property             (GObject         *object,
							 guint            param_id,
							 const GValue    *value,
							 GParamSpec      *pspec);
static void       gtk_dock_bar_get_property             (GObject         *object,
						 	 guint            param_id,
						 	 GValue          *value,
						 	 GParamSpec      *pspec);
static void       gtk_dock_bar_set_position             (GtkDockBar      *dock_bar,
							 GtkPositionType  new_position);
static void       gtk_dock_bar_real_orientation_changed (GtkDockBar      *dock_bar,
							 GtkOrientation   new_orientation);
static gboolean   gtk_dock_bar_button_changed           (GtkWidget       *widget,
							 GdkEventButton  *event);
static gboolean   gtk_dock_bar_motion                   (GtkWidget       *widget,
							 GdkEventMotion  *event);

/* GtkDockBox */

static void       gtk_dock_box_class_init               (GtkDockBoxClass *klass);
static void       gtk_dock_box_init                     (GtkDockBox      *box);

static gboolean   gtk_dock_box_widget_contains_point    (GtkWidget       *widget,
							 gint             x,
							 gint             y);
static GtkWidget* gtk_dock_box_get_widget_box           (GtkDockBox      *dock_box,
						 	 GtkPositionType  position);

static gpointer parent_class;
static const gchar *dock_box_key = "gtk-dock-bar-dock-box";
static guint dock_bar_signals[LAST_SIGNAL] = { 0 };

GtkType
gtk_dock_bar_get_type (void)
{
  static GtkType dock_bar_type = 0;

  if (!dock_bar_type)
    {
      static const GtkTypeInfo dock_bar_info =
      {
	"GtkDockBar",
	sizeof (GtkDockBar),
	sizeof (GtkDockBarClass),
	(GtkClassInitFunc) gtk_dock_bar_class_init,
	(GtkObjectInitFunc) gtk_dock_bar_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL,
      };
      dock_bar_type = gtk_type_unique (GTK_TYPE_HANDLE_BOX, &dock_bar_info);
    }

  return dock_bar_type;
}

static void
gtk_dock_bar_class_init (GtkDockBarClass *class)
{
  GObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GObjectClass *) class;
  widget_class = (GtkWidgetClass *) class;

  parent_class = g_type_class_peek_parent (class);

  object_class->set_property = gtk_dock_bar_set_property;
  object_class->get_property = gtk_dock_bar_get_property;

  g_object_class_install_property (object_class,
				   PROP_POSITION,
				   g_param_spec_enum ("position",
						      _("Position"),
						      _("The dock box edge to align the dock bar up against."),
						      GTK_TYPE_POSITION_TYPE,
						      GTK_POS_TOP,
						      G_PARAM_READABLE | G_PARAM_WRITABLE));

  widget_class->button_press_event = gtk_dock_bar_button_changed;
  widget_class->button_release_event = gtk_dock_bar_button_changed;
  widget_class->motion_notify_event = gtk_dock_bar_motion;

  class->orientation_changed = gtk_dock_bar_real_orientation_changed;
  class->position_changed = NULL;

  dock_bar_signals[ORIENTATION_CHANGED] =
    gtk_signal_new ("orientation_changed",
		    GTK_RUN_FIRST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkDockBarClass, orientation_changed),
		    g_cclosure_marshal_VOID__ENUM,
		    GTK_TYPE_NONE, 1,
		    GTK_TYPE_ORIENTATION);
  dock_bar_signals[POSITION_CHANGED] =
    gtk_signal_new ("position_changed",
		    GTK_RUN_FIRST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkDockBarClass, position_changed),
		    g_cclosure_marshal_VOID__ENUM,
		    GTK_TYPE_NONE, 1,
		    GTK_TYPE_POSITION_TYPE);
}

static void
gtk_dock_bar_init (GtkDockBar *dock_bar)
{
  dock_bar->position = GTK_POS_TOP;
  dock_bar->in_drag = FALSE;
}

static void
gtk_dock_bar_set_property (GObject      *object,
			   guint         prop_id,
			   const GValue *value,
			   GParamSpec   *pspec)
{
  GtkDockBar *dock_bar = GTK_DOCK_BAR (object);

  switch (prop_id)
    {
    case PROP_POSITION:
      gtk_dock_bar_set_position (dock_bar, g_value_get_enum (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void 
gtk_dock_bar_get_property (GObject    *object,
			     guint       prop_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  GtkDockBar *dock_bar = GTK_DOCK_BAR (object);

  switch (prop_id)
    {
    case PROP_POSITION:
      g_value_set_enum (value, dock_bar->position);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_dock_bar_new (GtkPositionType position)
{
  GtkDockBar *dock_bar;
  GtkPositionType handle_position;

  dock_bar = gtk_type_new (GTK_TYPE_DOCK_BAR);

  dock_bar->position = position;
  if (position == GTK_POS_TOP || position == GTK_POS_BOTTOM)
	handle_position = GTK_POS_LEFT;
  else
	handle_position = GTK_POS_TOP;
  gtk_handle_box_set_handle_position (GTK_HANDLE_BOX (dock_bar), handle_position);

  return GTK_WIDGET (dock_bar);
}

GtkOrientation
gtk_dock_bar_get_orientation (GtkDockBar *dock_bar)
{
  GtkOrientation orientation;

  if (dock_bar->position == GTK_POS_TOP || dock_bar->position == GTK_POS_BOTTOM)
	orientation = GTK_ORIENTATION_HORIZONTAL;
  else
	orientation = GTK_ORIENTATION_VERTICAL;

  return orientation;
}

GtkPositionType
gtk_dock_bar_get_position (GtkDockBar *dock_bar)
{
  g_return_val_if_fail (GTK_IS_DOCK_BAR (dock_bar), ((GtkPositionType)-1));

  return dock_bar->position;
}

static void
gtk_dock_bar_set_position (GtkDockBar      *dock_bar,
			   GtkPositionType  new_position)
{
  GtkOrientation old_orientation;
  GtkOrientation new_orientation;

  g_return_if_fail (GTK_IS_DOCK_BAR (dock_bar));

  if (dock_bar->position != new_position)
    {
      old_orientation = gtk_dock_bar_get_orientation (dock_bar);
      dock_bar->position = new_position;
      gtk_signal_emit (GTK_OBJECT (dock_bar),
		       dock_bar_signals[POSITION_CHANGED],
		       new_position);
      new_orientation = gtk_dock_bar_get_orientation (dock_bar);
      if (new_orientation != old_orientation)
        {
          gtk_signal_emit (GTK_OBJECT (dock_bar),
			   dock_bar_signals[ORIENTATION_CHANGED],
			   new_orientation);
        }
      g_object_notify (G_OBJECT (dock_bar), "position");
    }
}

static void
gtk_dock_bar_real_orientation_changed (GtkDockBar     *dock_bar,
					GtkOrientation  new_orientation)
{
  GtkPositionType handle_position;

  g_return_if_fail (GTK_IS_DOCK_BAR (dock_bar));

  if (new_orientation == GTK_ORIENTATION_HORIZONTAL)
	handle_position = GTK_POS_LEFT;
  else
	handle_position = GTK_POS_TOP;

  gtk_handle_box_set_handle_position (GTK_HANDLE_BOX (dock_bar), handle_position);
}

static gboolean
gtk_dock_bar_button_changed (GtkWidget      *widget,
			     GdkEventButton *event)
{
  GtkDockBar *dock_bar;
  gboolean event_handled;
  GdkCursor *fleur;

  dock_bar = GTK_DOCK_BAR (widget);

  event_handled = FALSE;
  if ((event->button == 1) && (event->type == GDK_BUTTON_PRESS || event->type == GDK_2BUTTON_PRESS))
    {
      GtkWidget *child;
      gboolean in_handle;

      child = GTK_BIN (dock_bar)->child;

      if (child)
	{
	  switch (GTK_HANDLE_BOX (dock_bar)->handle_position)
	    {
	    case GTK_POS_LEFT:
	      in_handle = event->x < DRAG_HANDLE_SIZE;
	      break;
	    case GTK_POS_TOP:
	      in_handle = event->y < DRAG_HANDLE_SIZE;
	      break;
	    case GTK_POS_RIGHT:
	      in_handle = event->x > 2 * GTK_CONTAINER (dock_bar)->border_width + child->allocation.width;
	      break;
	    case GTK_POS_BOTTOM:
	      in_handle = event->y > 2 * GTK_CONTAINER (dock_bar)->border_width + child->allocation.height;
	      break;
	    default:
	      in_handle = FALSE;
	      break;
	    }
	}
      else
        {
	  in_handle = FALSE;
	  event_handled = TRUE;
	}

      if (!in_handle)
        {
          dock_bar->in_drag = TRUE;
          fleur = gdk_cursor_new (GDK_FLEUR);
          if (gdk_pointer_grab (widget->window,
				FALSE,
				(GDK_BUTTON1_MOTION_MASK |
				GDK_POINTER_MOTION_HINT_MASK |
				GDK_BUTTON_RELEASE_MASK),
				NULL,
				fleur,
				GDK_CURRENT_TIME) != 0)
            {
              dock_bar->in_drag = FALSE;
            }
        }
      else if (GTK_WIDGET_CLASS (parent_class)->button_press_event)
        (* GTK_WIDGET_CLASS (parent_class)->button_press_event) (widget, event);
    }
    else if (event->type == GDK_BUTTON_RELEASE)
    {
      if (dock_bar->in_drag)
        {
	  gdk_pointer_ungrab (GDK_CURRENT_TIME);
	  dock_bar->in_drag = FALSE;
	  event_handled = TRUE;
	}
      else if (GTK_WIDGET_CLASS (parent_class)->button_release_event)
        (* GTK_WIDGET_CLASS (parent_class)->button_release_event) (widget, event);
    }

  return event_handled;
}

static gboolean
gtk_dock_bar_motion (GtkWidget      *widget,
		     GdkEventMotion *event)
{
  GtkDockBar *dock_bar;
  GtkDockBox *dock_box;
  GtkPositionType position;
  GtkAllocation a;
  gint x, y;
  gint dx, dy, y1, y2;
  GdkCursor *fleur;

  dock_bar = GTK_DOCK_BAR (widget);
  dock_box = g_object_get_data (G_OBJECT (widget), dock_box_key);

  if (!dock_bar->in_drag || !dock_box)
  {
    if (GTK_WIDGET_CLASS (parent_class)->motion_notify_event)
      (* GTK_WIDGET_CLASS (parent_class)->motion_notify_event) (widget, event);
    return FALSE;
  }

  position = gtk_dock_bar_get_position (dock_bar);
  a = GTK_WIDGET (dock_box)->allocation;
  gtk_widget_get_pointer (GTK_WIDGET (dock_box), &x, &y);
  x += a.x;
  y += a.y;

  /* If the cursor is inside one of the edge boxes just reorder the dock_bars. */
  if (!gtk_dock_box_widget_contains_point (dock_box->client_vbox, x, y))
    {

      GtkWidget *box;

      box = gtk_dock_box_get_widget_box (dock_box, position);

      if (box && gtk_dock_box_widget_contains_point (box, x, y))
        {
 	  GList *list;
	  gint i = 0;

	  for (list = GTK_BOX (box)->children; list; list = list->next)
	    {
	      GtkBoxChild *child_entry;

	      child_entry = list->data;

              if ((child_entry->widget != widget) &&
	          gtk_dock_box_widget_contains_point (child_entry->widget, x, y))
	        {
		  gtk_box_reorder_child (GTK_BOX (box), widget, i);
		  return TRUE;
		}
	      i++;
	    }
        }
      return TRUE;
    }

  /* Calculate which diagonal quadrant of the client_vbox the cursor is in. */
  dy = (gint)(((gfloat)(a.height) / (gfloat)(a.width)) * (x - a.x));
  y1 = a.y + dy;
  y2 = a.y + a.height - dy;

  /* Reduce cursor sensitivity by excluding the central area of the client_vbox. */
  dx = (a.width / 3);
  dy = (a.height / 3);
  a.x += dx;
  a.y += dy;
  a.width -= (dx * 2);
  a.height -= (dy * 2);
  
  if ((x >= a.x) && (x <= (a.x + a.width)) && (y >= a.y) && (y <= (a.y + a.height)))
    return TRUE;
    
  /* Set the dock bar positon based on the direction the float_window is being dragged. */
  if (y < y1)
    {
      if (y < y2)
	position = GTK_POS_TOP;
      else
	position = GTK_POS_RIGHT;
    }
  else
    {
      if (y < y2)
	position = GTK_POS_LEFT;
      else
	position = GTK_POS_BOTTOM;
    }

  /* If the position has changed move the dock_bar */
  if (position != gtk_dock_bar_get_position (dock_bar))
    {
      gdk_pointer_ungrab (GDK_CURRENT_TIME);
      gtk_dock_box_move_dock_bar (dock_box, dock_bar, position);
      fleur = gdk_cursor_new (GDK_FLEUR);
      if (gdk_pointer_grab (widget->window,
			    FALSE,
			    (GDK_BUTTON1_MOTION_MASK |
			     GDK_POINTER_MOTION_HINT_MASK |
			     GDK_BUTTON_RELEASE_MASK),
			    NULL,
			    fleur,
			    GDK_CURRENT_TIME) != 0)
        {
          dock_bar->in_drag = FALSE;
        }
    }
  return TRUE;
}

/* GtkDockBox */

GtkType
gtk_dock_box_get_type (void)
{
  static GtkType dock_box_type = 0;

  if (!dock_box_type)
    {
      static const GtkTypeInfo dock_box_info =
      {
	"GtkDockBox",
	sizeof (GtkDockBox),
	sizeof (GtkDockBoxClass),
	(GtkClassInitFunc) gtk_dock_box_class_init,
	(GtkObjectInitFunc) gtk_dock_box_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL,
      };
      dock_box_type = gtk_type_unique (GTK_TYPE_VBOX, &dock_box_info);
    }

  return dock_box_type;
}

static void
gtk_dock_box_class_init (GtkDockBoxClass *class)
{
}

static void
gtk_dock_box_init (GtkDockBox *dock_box)
{
  dock_box->top_vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dock_box), dock_box->top_vbox, FALSE, FALSE, 0);
  dock_box->hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dock_box), dock_box->hbox, TRUE, TRUE, 0);
  dock_box->left_hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dock_box->hbox), dock_box->left_hbox, FALSE, FALSE, 0);
  dock_box->client_vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dock_box->hbox), dock_box->client_vbox, TRUE, TRUE, 0);
  dock_box->right_hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dock_box->hbox), dock_box->right_hbox, FALSE, FALSE, 0);
  dock_box->bottom_vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dock_box), dock_box->bottom_vbox, FALSE, FALSE, 0);
  gtk_widget_show (dock_box->top_vbox);
  gtk_widget_show_all (dock_box->hbox);
  gtk_widget_show (dock_box->bottom_vbox);
}

GtkWidget*
gtk_dock_box_new (void)
{
  return GTK_WIDGET (gtk_type_new (GTK_TYPE_DOCK_BOX));
}

void
gtk_dock_box_add_dock_bar (GtkDockBox *dock_box,
			   GtkDockBar *dock_bar)
{
  GtkWidget *widget;
  GtkPositionType position;

  g_return_if_fail (GTK_IS_DOCK_BOX (dock_box));
  g_return_if_fail (GTK_IS_DOCK_BAR (dock_bar));

  widget = GTK_WIDGET (dock_bar);
  position = gtk_dock_bar_get_position (dock_bar);

  switch (position)
    {
    case GTK_POS_LEFT:
      gtk_box_pack_start (GTK_BOX (dock_box->left_hbox), widget, FALSE, FALSE, 0);
      break;

    case GTK_POS_TOP:
      gtk_box_pack_start (GTK_BOX (dock_box->top_vbox), widget, FALSE, FALSE, 0);
      break;

    case GTK_POS_RIGHT:
      gtk_box_pack_start (GTK_BOX (dock_box->right_hbox), widget, FALSE, FALSE, 0);
      break;

    case GTK_POS_BOTTOM:
      gtk_box_pack_start (GTK_BOX (dock_box->bottom_vbox), widget, FALSE, FALSE, 0);
      break;
    }
    
  g_object_set_data (G_OBJECT (dock_bar), dock_box_key, dock_box);
  gtk_widget_show (widget);
}

void       
gtk_dock_box_move_dock_bar (GtkDockBox      *dock_box,
			    GtkDockBar      *dock_bar,
			    GtkPositionType  new_position)
{
  GtkWidget *widget;
  GtkPositionType position;

  g_return_if_fail (GTK_IS_DOCK_BOX (dock_box));
  g_return_if_fail (GTK_IS_DOCK_BAR (dock_bar));

  widget = GTK_WIDGET (dock_bar);
  position = gtk_dock_bar_get_position (dock_bar);

  if (new_position != position)
  {
    gtk_widget_hide (widget);
    gtk_dock_bar_set_position (dock_bar, new_position);
    
    switch (new_position)
      {
      case GTK_POS_LEFT:
        gtk_widget_reparent (widget, GTK_WIDGET (dock_box->left_hbox));
	gtk_box_set_child_packing (GTK_BOX (dock_box->left_hbox), widget, FALSE, FALSE, 0, GTK_PACK_START);
	break;

      case GTK_POS_TOP:
        gtk_widget_reparent (widget, GTK_WIDGET (dock_box->top_vbox));
	gtk_box_set_child_packing (GTK_BOX (dock_box->top_vbox), widget, FALSE, FALSE, 0, GTK_PACK_START);
	break;

      case GTK_POS_RIGHT:
        gtk_widget_reparent (widget, GTK_WIDGET (dock_box->right_hbox));
	gtk_box_set_child_packing (GTK_BOX (dock_box->right_hbox), widget, FALSE, FALSE, 0, GTK_PACK_END);
	break;

      case GTK_POS_BOTTOM:
        gtk_widget_reparent (widget, GTK_WIDGET (dock_box->bottom_vbox));
	gtk_box_set_child_packing (GTK_BOX (dock_box->bottom_vbox), widget, FALSE, FALSE, 0, GTK_PACK_END);
	break;
      }
      
    gtk_handle_box_set_snap_edge (GTK_HANDLE_BOX (dock_bar), new_position);
    gtk_widget_show (widget);
  }
}

void
gtk_dock_box_remove_dock_bar (GtkDockBox *dock_box,
			      GtkDockBar *dock_bar)
{
  GtkPositionType position;
  GtkWidget *box;

  g_return_if_fail (GTK_IS_DOCK_BOX (dock_box));
  g_return_if_fail (GTK_IS_DOCK_BAR (dock_bar));

  position = gtk_dock_bar_get_position (dock_bar);
  box = gtk_dock_box_get_widget_box (dock_box, position);
  gtk_container_remove (GTK_CONTAINER (box), GTK_WIDGET (dock_bar));
  g_object_set_data (G_OBJECT (dock_bar), dock_box_key, NULL);
}

gint       
gtk_dock_box_get_dock_bar_index (GtkDockBox *dock_box,
				 GtkDockBar *dock_bar)
{
  GtkWidget *widget;
  GtkPositionType position;
  GtkWidget *box;
  gint index = -1;

  g_return_val_if_fail (GTK_IS_DOCK_BOX (dock_box), index);
  g_return_val_if_fail (GTK_IS_DOCK_BAR (dock_bar), index);

  widget = GTK_WIDGET (dock_bar);
  position = gtk_dock_bar_get_position (dock_bar);
  box = gtk_dock_box_get_widget_box (dock_box, position);

  if (box)
    {
      GtkPackType pack_type;
      GList *children;
      GList *next;
      
      gtk_box_query_child_packing (GTK_BOX (dock_box), 
				   GTK_WIDGET (dock_bar),
				   NULL, 
				   NULL, 
				   NULL,
				   &pack_type);

      children = GTK_BOX (box)->children;
      next = (pack_type == GTK_PACK_START) ? children : g_list_last(children);

      while (next != 0)
        {
	  GtkBoxChild* child_entry = ((GtkBoxChild*)next->data);
	  if (child_entry->pack == (guint)pack_type)
            {
	      index++;
	      if (child_entry->widget == widget)
	        break;
	    }
	  next = (pack_type == GTK_PACK_START) ? g_list_next(next) : g_list_previous(next);
	}

    }

  return index;
}

static gboolean
gtk_dock_box_widget_contains_point (GtkWidget *widget,
				    gint       x,
				    gint       y)
{
  GtkAllocation a;
  gboolean contains;

  a = widget->allocation;
  contains = ((x >= a.x) && (x <= (a.x + a.width)) && (y >= a.y) && (y <= (a.y + a.height)));

  return contains;
}

static GtkWidget*
gtk_dock_box_get_widget_box (GtkDockBox      *dock_box,
			     GtkPositionType  position)
{
  GtkWidget *box = NULL;

  switch (position)
    {
    case GTK_POS_LEFT:
      box = dock_box->left_hbox;
      break;

    case GTK_POS_TOP:
      box = dock_box->top_vbox;
      break;

    case GTK_POS_RIGHT:
      box = dock_box->right_hbox;
      break;

    case GTK_POS_BOTTOM:
      box = dock_box->bottom_vbox;
      break;
    }

  return box;
}

