/* GTK - The GIMP Toolkit
 * Copyright (C) 2002 Jeff Franks <jcf@tpg.com.au>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
/* GtkDockBar and GtkDockBox together form a basic dockable widget. Initialy I wrote
 * this in C++ but thought it might serve as a starting point from which to develope
 * a GTK+-2.X dockable widget. Apart from the drag and dock code they add very little
 * new code. Instead they build on the widget code GTK already provides. Any feedback
 * would be welcome.
 *
 * GtkDockBar derives from GtkHandleBox and adds to it the concept of a docking edge. 
 * A GtkDockBar remembers it's docked edge and emits a 'position_changed_signal' when
 * it changes and an 'orientation_changed_signal' when it's orientation changes. You 
 * drag and dock a GtkDockBar by clicking and holding the left mouse button down somewhere
 * inside its's boundaries. If the cursor is inside the handle you get the default
 * GtkHandleBox drag and dock behaviour. If the cursor is outside the handle but still
 * inside the dock bar you can drag and dock the GtkDockBar to any one of the four edges
 * of it's parent GtkDockBox: left, top, right or bottom. When there is more than one 
 * GtkDockBar docked to an edge you can drag and dock the dock bars to reorder their
 * docked positions.
 *
 * GtkDockBox is a composite widget of horizontal and vertical boxes and derives from
 * GtkVBox. It is the dock bar manager and is responsible correctly moving the dock bars
 * around it's four edges. The layout of the DockBox looks like this:
 *       ___________________________________________________________________________
 *      |                               GtkDockBox                                  |
 *      |  _______________________________________________________________________  |
 *      | |                                                                       | |
 *      | |                               top_vbox                                | |
 *      | |_______________________________________________________________________| |
 *      |  ________   _________________________________________________   ________  |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |  left  | |                                                 | |  right | |
 *      | |  hbox  | |                   client_vbox                   | |  hbox  | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |        | |                                                 | |        | |
 *      | |________| |_________________________________________________| |________| |
 *      |  _______________________________________________________________________  |
 *      | |                                                                       | |
 *      | |                              bottom_vbox                              | |
 *      | |_______________________________________________________________________| |
 *      |___________________________________________________________________________|
 *
 * When you drag a GtkDockBar and dock it to one of the four edges of a GtkDockBox
 * the dock bar is reparented and packed into one of the four edge boxes: left_hbox,
 * top_vbox, right_hbox and bottom_vbox. The client_vbox is the client area of the
 * GtkDockBox and is the box that you pack your widgets into.
 *
 * During a drag the GtkDockBar uses the following client_vbox layout to decide which
 * edge it should be docked to when the left mouse button is released:
 *       _______________________________
 *      |\                             /|
 *      | \       "client_vbox"       / |
 *      |  \                         /  |
 *      |   \          top          /   |
 *      |    \                     /    |
 *      |     \                   /     |
 *      |      \                 /      |
 *      |       \_______________/       |
 *      |       |               |       |
 *      |       |               |       |
 *      | left  |   no change   | right |
 *      |       |               |       |
 *      |       |_______________|       |
 *      |       /               \       |
 *      |      /                 \      |
 *      |     /                   \     |
 *      |    /                     \    |
 *      |   /         bottom        \   |
 *      |  /                         \  |
 *      | /                           \ |
 *      |/_____________________________\|
 *
 * The layout is based on two diagonal lines dividing the client_vbox into four
 * quadrants. When the cursor is dragged into one of these four quadrants: left,
 * top, right and bottom, that is the edge the dock bar is docked to when the left
 * mouse is released. The central area of the client_vbox is excluded from the
 * decision to minimize cursor sensitivity.
 *
 * The dock bar positions are easy to restore. Two pieces of information are
 * required, the docked edge's GtkPositionType value and the GtkDockBar's zero 
 * based index on it's docked edge. Both these values can be saved in a user
 * preference file and reloaded when the application is next run. The GtkDockBar
 * method gtk_dock_bar_get_position() returns the GtkPositonType value and the
 * GtkDockBox gtk_dock_box_get_dock_bar_index() method returns the zero-based
 * index.
 */

#ifndef __GTK_DOCK_BOX_H__
#define __GTK_DOCK_BOX_H__


#include <gdk/gdk.h>
#include <gtk/gtkbin.h>
#include <gtk/gtkhandlebox.h>
#include <gtk/gtkvbox.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* GtkDockBar */

#define GTK_TYPE_DOCK_BAR            (gtk_dock_bar_get_type ())
#define GTK_DOCK_BAR(obj)            (GTK_CHECK_CAST ((obj), GTK_TYPE_DOCK_BAR, GtkDockBar))
#define GTK_DOCK_BAR_CLASS(klass)    (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_DOCK_BAR, GtkDockBarClass))
#define GTK_IS_DOCK_BAR(obj)         (GTK_CHECK_TYPE ((obj), GTK_TYPE_DOCK_BAR))
#define GTK_IS_DOCK_BAR_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_DOCK_BAR))
#define GTK_DOCK_BAR_GET_CLASS(obj)  (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_DOCK_BAR, GtkDockBarClass))

typedef struct _GtkDockBar	 GtkDockBar;
typedef struct _GtkDockBarClass  GtkDockBarClass;

struct _GtkDockBar
{
  GtkHandleBox handle_box;

  /*< private >*/
  GtkPositionType position;
  guint           in_drag : 1;
};

struct _GtkDockBarClass
{
  GtkHandleBoxClass parent_class;
  
  void (*orientation_changed) (GtkDockBar *dock_bar, GtkOrientation new_orientation);

  void (*position_changed)    (GtkDockBar *dock_bar, GtkPositionType new_position);
};


GtkType	        gtk_dock_bar_get_type (void) G_GNUC_CONST;
GtkWidget*      gtk_dock_bar_new      (GtkPositionType position);

GtkOrientation  gtk_dock_bar_get_orientation (GtkDockBar *dock_bar);
GtkPositionType gtk_dock_bar_get_position    (GtkDockBar *dock_bar);
                                              
/* GtkDockBox */

#define GTK_TYPE_DOCK_BOX            (gtk_dock_box_get_type ())
#define GTK_DOCK_BOX(obj)            (GTK_CHECK_CAST ((obj), GTK_TYPE_DOCK_BOX, GtkDockBox))
#define GTK_DOCK_BOX_CLASS(klass)    (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_DOCK_BOX, GtkDockBoxClass))
#define GTK_IS_DOCK_BOX(obj)         (GTK_CHECK_TYPE ((obj), GTK_TYPE_DOCK_BOX))
#define GTK_IS_DOCK_BOX_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_DOCK_BOX))
#define GTK_DOCK_BOX_GET_CLASS(obj)  (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_DOCK_BOX, GtkDockBoxClass))


typedef struct _GtkDockBox	 GtkDockBox;
typedef struct _GtkDockBoxClass  GtkDockBoxClass;

struct _GtkDockBox
{
  GtkVBox vbox;
  
  GtkWidget *client_vbox;
  
  /*< private >*/
  GtkWidget *hbox;
  GtkWidget *top_vbox;
  GtkWidget *left_hbox;
  GtkWidget *right_hbox;
  GtkWidget *bottom_vbox;
};

struct _GtkDockBoxClass
{
  GtkVBoxClass parent_class;
};


GtkType	   gtk_dock_box_get_type (void) G_GNUC_CONST;
GtkWidget* gtk_dock_box_new	 (void);

void       gtk_dock_box_add_dock_bar       (GtkDockBox *dock_box,
                                            GtkDockBar *dock_bar);
void       gtk_dock_box_move_dock_bar      (GtkDockBox *dock_box,
                                            GtkDockBar *dock_bar,
                                            GtkPositionType new_position);
void       gtk_dock_box_remove_dock_bar    (GtkDockBox *dock_box,
                                            GtkDockBar *dock_bar);
gint       gtk_dock_box_get_dock_bar_index (GtkDockBox *dock_box,
                                            GtkDockBar *dock_bar);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GTK_DOCK_BOX_H__ */
